"""
**NetRef**: a transparent *network reference*. This module contains quite a lot
of *magic*, so beware.
"""
import sys
import inspect
import types
from ..lib.compat import pickle, is_py3k, maxint
from ..core import consts


_local_netref_attrs = frozenset([
    '____conn__', '____oid__', '__class__', '__cmp__', '__del__', '__delattr__',
    '__dir__', '__doc__', '__getattr__', '__getattribute__', '__hash__',
    '__init__', '__metaclass__', '__module__', '__new__', '__reduce__',
    '__reduce_ex__', '__repr__', '__setattr__', '__slots__', '__str__',
    '__weakref__', '__dict__', '__members__', '__methods__',
])
"""the set of attributes that are local to the netref object"""

_builtin_types = [
    type, object, bool, complex, dict, float, int, list, slice, str, tuple, set,
    frozenset, Exception, type(None), types.BuiltinFunctionType, types.GeneratorType,
    types.MethodType, types.CodeType, types.FrameType, types.TracebackType,
    types.ModuleType, types.FunctionType,

    type(int.__add__),      # wrapper_descriptor
    type((1).__add__),      # method-wrapper
    type(iter([])),         # listiterator
    type(iter(())),         # tupleiterator
    type(iter(set())),      # setiterator
]
"""a list of types considered built-in (shared between connections)"""

try:
    BaseException
except NameError:
    pass
else:
    _builtin_types.append(BaseException)

if is_py3k:
    _builtin_types.extend([
        bytes, bytearray, type(iter(range(10))), memoryview,
    ])
else:
    _builtin_types.extend([
        basestring, unicode, long, xrange, type(iter(xrange(10))), file,
        types.InstanceType, types.ClassType, types.DictProxyType,
    ])

_normalized_builtin_types = dict(((t.__name__, t.__module__), t)
    for t in _builtin_types)

def syncreq(proxy, handler, *args):
    """Performs a synchronous request on the given proxy object.
    Not intended to be invoked directly.

    :param proxy: the proxy on which to issue the request
    :param handler: the request handler (one of the ``HANDLE_XXX`` members of
                    ``rpyc.protocol.consts``)
    :param args: arguments to the handler

    :raises: any exception raised by the operation will be raised
    :returns: the result of the operation
    """
    conn = object.__getattribute__(proxy, "____conn__")()
    if not conn:
        raise ReferenceError('weakly-referenced object no longer exists')
    oid = object.__getattribute__(proxy, "____oid__")
    return conn.sync_request(handler, oid, *args)

def asyncreq(proxy, handler, *args):
    """Performs an asynchronous request on the given proxy object.
    Not intended to be invoked directly.

    :param proxy: the proxy on which to issue the request
    :param handler: the request handler (one of the ``HANDLE_XXX`` members of
                    ``rpyc.protocol.consts``)
    :param args: arguments to the handler

    :returns: an :class:`AsyncResult <rpyc.core.async.AsyncResult>` representing
              the operation
    """
    conn = object.__getattribute__(proxy, "____conn__")()
    if not conn:
        raise ReferenceError('weakly-referenced object no longer exists')
    oid = object.__getattribute__(proxy, "____oid__")
    return conn.async_request(handler, oid, *args)

class NetrefMetaclass(type):
    """A *metaclass* used to customize the ``__repr__`` of ``netref`` classes.
    It is quite useless, but it makes debugging and interactive programming
    easier"""

    __slots__ = ()
    def __repr__(self):
        if self.__module__:
            return "<netref class '%s.%s'>" % (self.__module__, self.__name__)
        else:
            return "<netref class '%s'>" % (self.__name__,)

class BaseNetref(object):
    """The base netref class, from which all netref classes derive. Some netref
    classes are "pre-generated" and cached upon importing this module (those
    defined in the :data:`_builtin_types`), and they are shared between all
    connections.

    The rest of the netref classes are created by :meth:`rpyc.core.protocl.Connection._unbox`,
    and are private to the connection.

    Do not use this class directly; use :func:`class_factory` instead.

    :param conn: the :class:`rpyc.core.protocol.Connection` instance
    :param oid: the unique object ID of the remote object
    """
    # this is okay with py3k -- see below
    __metaclass__ = NetrefMetaclass
    __slots__ = ["____conn__", "____oid__", "__weakref__"]
    def __init__(self, conn, oid):
        self.____conn__ = conn
        self.____oid__ = oid
    def __del__(self):
        try:
            asyncreq(self, consts.HANDLE_DEL)
        except Exception:
            # raised in a destructor, most likely on program termination,
            # when the connection might have already been closed.
            # it's safe to ignore all exceptions here
            pass

    def __getattribute__(self, name):
        if name in _local_netref_attrs:
            if name == "__class__":
                cls = object.__getattribute__(self, "__class__")
                if cls is None:
                    cls = self.__getattr__("__class__")
                return cls
            elif name == "__doc__":
                return self.__getattr__("__doc__")
            elif name == "__members__":                       # for Python < 2.6
                return self.__dir__()
            else:
                return object.__getattribute__(self, name)
        elif name == "__call__":                          # IronPython issue #10
            return object.__getattribute__(self, "__call__")
        else:
            return syncreq(self, consts.HANDLE_GETATTR, name)
    def __getattr__(self, name):
        return syncreq(self, consts.HANDLE_GETATTR, name)
    def __delattr__(self, name):
        if name in _local_netref_attrs:
            object.__delattr__(self, name)
        else:
            syncreq(self, consts.HANDLE_DELATTR, name)
    def __setattr__(self, name, value):
        if name in _local_netref_attrs:
            object.__setattr__(self, name, value)
        else:
            syncreq(self, consts.HANDLE_SETATTR, name, value)
    def __dir__(self):
        return list(syncreq(self, consts.HANDLE_DIR))

    # support for metaclasses
    def __hash__(self):
        return syncreq(self, consts.HANDLE_HASH)
    def __cmp__(self, other):
        return syncreq(self, consts.HANDLE_CMP, other)
    def __repr__(self):
        return syncreq(self, consts.HANDLE_REPR)
    def __str__(self):
        return syncreq(self, consts.HANDLE_STR)

    # support for pickling netrefs
    def __reduce_ex__(self, proto):
        return pickle.loads, (syncreq(self, consts.HANDLE_PICKLE, proto),)

if not isinstance(BaseNetref, NetrefMetaclass):
    # python 2 and 3 compatible metaclass...
    ns = dict(BaseNetref.__dict__)
    for slot in BaseNetref.__slots__:
        ns.pop(slot)
    BaseNetref = NetrefMetaclass(BaseNetref.__name__, BaseNetref.__bases__, ns)


def _make_method(name, doc):
    """creates a method with the given name and docstring that invokes
    :func:`syncreq` on its `self` argument"""

    slicers = {"__getslice__" : "__getitem__", "__delslice__" : "__delitem__", "__setslice__" : "__setitem__"}

    name = str(name)                                      # IronPython issue #10
    if name == "__call__":
        def __call__(_self, *args, **kwargs):
            kwargs = tuple(kwargs.items())
            return syncreq(_self, consts.HANDLE_CALL, args, kwargs)
        __call__.__doc__ = doc
        return __call__
    elif name in slicers:                                 # 32/64 bit issue #41
        def method(self, start, stop, *args):
            if stop == maxint:
                stop = None
            return syncreq(self, consts.HANDLE_OLDSLICING, slicers[name], name, start, stop, args)
        method.__name__ = name
        method.__doc__ = doc
        return method
    else:
        def method(_self, *args, **kwargs):
            kwargs = tuple(kwargs.items())
            return syncreq(_self, consts.HANDLE_CALLATTR, name, args, kwargs)
        method.__name__ = name
        method.__doc__ = doc
        return method

def inspect_methods(obj):
    """introspects the given (local) object, returning a list of all of its
    methods (going up the MRO).

    :param obj: any local (not proxy) python object

    :returns: a list of ``(method name, docstring)`` tuples of all the methods
              of the given object
    """
    methods = {}
    attrs = {}
    if isinstance(obj, type):
        # don't forget the darn metaclass
        mros = list(reversed(type(obj).__mro__)) + list(reversed(obj.__mro__))
    else:
        mros = reversed(type(obj).__mro__)
    for basecls in mros:
        attrs.update(basecls.__dict__)
    for name, attr in attrs.items():
        if name not in _local_netref_attrs and hasattr(attr, "__call__"):
            methods[name] = inspect.getdoc(attr)
    return methods.items()

def class_factory(clsname, modname, methods):
    """Creates a netref class proxying the given class

    :param clsname: the class's name
    :param modname: the class's module name
    :param methods: a list of ``(method name, docstring)`` tuples, of the methods
                    that the class defines

    :returns: a netref class
    """
    clsname = str(clsname)                                # IronPython issue #10
    modname = str(modname)                                # IronPython issue #10
    ns = {"__slots__" : ()}
    for name, doc in methods:
        name = str(name)                                  # IronPython issue #10
        if name not in _local_netref_attrs:
            ns[name] = _make_method(name, doc)
    ns["__module__"] = modname
    if modname in sys.modules and hasattr(sys.modules[modname], clsname):
        ns["__class__"] = getattr(sys.modules[modname], clsname)
    elif (clsname, modname) in _normalized_builtin_types:
        ns["__class__"] = _normalized_builtin_types[clsname, modname]
    else:
        # to be resolved by the instance
        ns["__class__"] = None
    return type(clsname, (BaseNetref,), ns)

builtin_classes_cache = {}
"""The cache of built-in netref classes (each of the types listed in
:data:`_builtin_types`). These are shared between all RPyC connections"""

# init the builtin_classes_cache
for cls in _builtin_types:
    builtin_classes_cache[cls.__name__, cls.__module__] = class_factory(
        cls.__name__, cls.__module__, inspect_methods(cls))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2013- Yan Shoshitaishvili aka. zardus
#                     Ruoyu Wang aka. fish
#                     Andrew Dutcher aka. rhelmot
#                     Kevin Borgolte aka. cao
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# :note: RemoteIDALink and get_memory must only be exported for rpyc

from .idalink import MODULE_DIR, IDA_DIR, LOGFILE, idalink, RemoteIDALink
from .memory import get_memory

__all__ = ["MODULE_DIR", "IDA_DIR", "LOGFILE", "idalink", "RemoteIDALink",
           "get_memory"]



import sys
import os
import inspect
from ..lib.compat import pickle
from .. import SlaveService
from ..utils import factory


DEFAULT_SERVER_PORT = 18812
DEFAULT_SERVER_SSL_PORT = 18821


#===============================================================================
# connecting
#===============================================================================
def connect_channel(channel):
    """
    Creates an RPyC connection over the given ``channel``

    :param channel: the :class:`rpyc.core.channel.Channel` instance

    :returns: an RPyC connection exposing ``SlaveService``
    """
    return factory.connect_channel(channel, SlaveService)

def connect_stream(stream):
    """
    Creates an RPyC connection over the given stream

    :param channel: the :class:`rpyc.core.stream.Stream` instance

    :returns: an RPyC connection exposing ``SlaveService``
    """
    return factory.connect_stream(stream, SlaveService)

def connect_stdpipes():
    """
    Creates an RPyC connection over the standard pipes (``stdin`` and ``stdout``)

    :returns: an RPyC connection exposing ``SlaveService``
    """
    return factory.connect_stdpipes(SlaveService)

def connect_pipes(input, output):
    """
    Creates an RPyC connection over two pipes

    :param input: the input pipe
    :param output: the output pipe

    :returns: an RPyC connection exposing ``SlaveService``
    """
    return factory.connect_pipes(input, output, SlaveService)

def connect(host, port = DEFAULT_SERVER_PORT, ipv6 = False):
    """
    Creates a socket connection to the given host and port.

    :param host: the host to connect to
    :param port: the TCP port
    :param ipv6: whether to create an IPv6 socket or IPv4

    :returns: an RPyC connection exposing ``SlaveService``
    """
    return factory.connect(host, port, SlaveService, ipv6 = ipv6)

def unix_connect(path):
    """
    Creates a socket connection to the given host and port.

    :param path: the path to the unix domain socket

    :returns: an RPyC connection exposing ``SlaveService``
    """
    return factory.unix_connect(path, SlaveService)

def ssl_connect(host, port = DEFAULT_SERVER_SSL_PORT, keyfile = None,
        certfile = None, ca_certs = None, cert_reqs = None, ssl_version = None,
        ciphers = None, ipv6 = False):
    """Creates a secure (``SSL``) socket connection to the given host and port,
    authenticating with the given certfile and CA file.

    :param host: the host to connect to
    :param port: the TCP port to use
    :param ipv6: whether to create an IPv6 socket or an IPv4 one

    The following arguments are passed directly to
    `ssl.wrap_socket <http://docs.python.org/dev/library/ssl.html#ssl.wrap_socket>`_:

    :param keyfile: see ``ssl.wrap_socket``. May be ``None``
    :param certfile: see ``ssl.wrap_socket``. May be ``None``
    :param ca_certs: see ``ssl.wrap_socket``. May be ``None``
    :param cert_reqs: see ``ssl.wrap_socket``. By default, if ``ca_cert`` is specified,
                      the requirement is set to ``CERT_REQUIRED``; otherwise it is
                      set to ``CERT_NONE``
    :param ssl_version: see ``ssl.wrap_socket``. The default is ``PROTOCOL_TLSv1``
    :param ciphers: see ``ssl.wrap_socket``. May be ``None``. New in Python 2.7/3.2

    :returns: an RPyC connection exposing ``SlaveService``

    .. _wrap_socket:
    """
    return factory.ssl_connect(host, port, keyfile = keyfile, certfile = certfile,
        ssl_version = ssl_version, ca_certs = ca_certs, service = SlaveService,
        ipv6 = ipv6)

def ssh_connect(sshctx, remote_port):
    """Connects to the remote server over an SSH tunnel. See
    :func:`rpyc.utils.factory.ssh_connect`.

    :param sshctx: the :class:`rpyc.utils.ssh.SshContext` instance
    :param remote_port: the remote TCP port

    :returns: an RPyC connection exposing ``SlaveService``
    """
    return factory.ssh_connect(sshctx, remote_port, SlaveService)

def connect_subproc(server_file = None):
    """Runs an RPyC classic server as a subprocess and returns an RPyC
    connection to it over stdio

    :param server_file: The full path to the server script (``rpyc_classic.py``).
                        If not given, ``which rpyc_classic.py`` will be attempted.

    :returns: an RPyC connection exposing ``SlaveService``
    """
    if server_file is None:
        server_file = os.popen("which rpyc_classic.py").read().strip()
        if not server_file:
            raise ValueError("server_file not given and could not be inferred")
    return factory.connect_subproc([sys.executable, "-u", server_file, "-q", "-m", "stdio"],
        SlaveService)

def connect_thread():
    """
    Starts a SlaveService on a thread and connects to it. Useful for testing
    purposes. See :func:`rpyc.utils.factory.connect_thread`

    :returns: an RPyC connection exposing ``SlaveService``
    """
    return factory.connect_thread(SlaveService, remote_service = SlaveService)

def connect_multiprocess(args = {}):
    """
    Starts a SlaveService on a multiprocess process and connects to it.
    Useful for testing purposes and running multicore code thats uses shared
    memory. See :func:`rpyc.utils.factory.connect_multiprocess`

    :returns: an RPyC connection exposing ``SlaveService``
    """
    return factory.connect_multiprocess(SlaveService, remote_service = SlaveService, args=args)


#===============================================================================
# remoting utilities
#===============================================================================

def upload(conn, localpath, remotepath, filter = None, ignore_invalid = False, chunk_size = 16000):
    """uploads a file or a directory to the given remote path

    :param localpath: the local file or directory
    :param remotepath: the remote path
    :param filter: a predicate that accepts the filename and determines whether
                   it should be uploaded; None means any file
    :param chunk_size: the IO chunk size
    """
    if os.path.isdir(localpath):
        upload_dir(conn, localpath, remotepath, filter, chunk_size)
    elif os.path.isfile(localpath):
        upload_file(conn, localpath, remotepath, chunk_size)
    else:
        if not ignore_invalid:
            raise ValueError("cannot upload %r" % (localpath,))

def upload_file(conn, localpath, remotepath, chunk_size = 16000):
    lf = open(localpath, "rb")
    rf = conn.builtin.open(remotepath, "wb")
    while True:
        buf = lf.read(chunk_size)
        if not buf:
            break
        rf.write(buf)
    lf.close()
    rf.close()

def upload_dir(conn, localpath, remotepath, filter = None, chunk_size = 16000):
    if not conn.modules.os.path.isdir(remotepath):
        conn.modules.os.makedirs(remotepath)
    for fn in os.listdir(localpath):
        if not filter or filter(fn):
            lfn = os.path.join(localpath, fn)
            rfn = conn.modules.os.path.join(remotepath, fn)
            upload(conn, lfn, rfn, filter = filter, ignore_invalid = True, chunk_size = chunk_size)

def download(conn, remotepath, localpath, filter = None, ignore_invalid = False, chunk_size = 16000):
    """
    download a file or a directory to the given remote path

    :param localpath: the local file or directory
    :param remotepath: the remote path
    :param filter: a predicate that accepts the filename and determines whether
                   it should be downloaded; None means any file
    :param chunk_size: the IO chunk size
    """
    if conn.modules.os.path.isdir(remotepath):
        download_dir(conn, remotepath, localpath, filter)
    elif conn.modules.os.path.isfile(remotepath):
        download_file(conn, remotepath, localpath, chunk_size)
    else:
        if not ignore_invalid:
            raise ValueError("cannot download %r" % (remotepath,))

def download_file(conn, remotepath, localpath, chunk_size = 16000):
    rf = conn.builtin.open(remotepath, "rb")
    lf = open(localpath, "wb")
    while True:
        buf = rf.read(chunk_size)
        if not buf:
            break
        lf.write(buf)
    lf.close()
    rf.close()

def download_dir(conn, remotepath, localpath, filter = None, chunk_size = 16000):
    if not os.path.isdir(localpath):
        os.makedirs(localpath)
    for fn in conn.modules.os.listdir(remotepath):
        if not filter or filter(fn):
            rfn = conn.modules.os.path.join(remotepath, fn)
            lfn = os.path.join(localpath, fn)
            download(conn, rfn, lfn, filter = filter, ignore_invalid = True)

def upload_package(conn, module, remotepath = None, chunk_size = 16000):
    """
    uploads a module or a package to the remote party

    :param conn: the RPyC connection to use
    :param module: the local module/package object to upload
    :param remotepath: the remote path (if ``None``, will default to the
                       remote system's python library (as reported by
                       ``distutils``)
    :param chunk_size: the IO chunk size

    .. note:: ``upload_module`` is just an alias to ``upload_package``

    example::

       import foo.bar
       ...
       rpyc.classic.upload_package(conn, foo.bar)

    """
    if remotepath is None:
        site = conn.modules["distutils.sysconfig"].get_python_lib()
        remotepath = conn.modules.os.path.join(site, module.__name__)
    localpath = os.path.dirname(inspect.getsourcefile(module))
    upload(conn, localpath, remotepath, chunk_size = chunk_size)

upload_module = upload_package

def obtain(proxy):
    """obtains (copies) a remote object from a proxy object. the object is
    ``pickled`` on the remote side and ``unpickled`` locally, thus moved
    **by value**. changes made to the local object will not reflect remotely.

    :param proxy: an RPyC proxy object

    .. note:: the remote object to must be ``pickle``-able

    :returns: a copy of the remote object
    """
    return pickle.loads(pickle.dumps(proxy))

def deliver(conn, localobj):
    """delivers (recreates) a local object on the other party. the object is
    ``pickled`` locally and ``unpickled`` on the remote side, thus moved
    **by value**. changes made to the remote object will not reflect locally.

    :param conn: the RPyC connection
    :param localobj: the local object to deliver

    .. note:: the object must be ``picklable``

    :returns: a proxy to the remote object
    """
    return conn.modules["rpyc.lib.compat"].pickle.loads(pickle.dumps(localobj))

class redirected_stdio(object):
    """
    Redirects the other party's ``stdin``, ``stdout`` and ``stderr`` to
    those of the local party, so remote IO will occur locally. It was
    originally written as a ``contextmanager``, but was turned into a class
    for compatibility with python 2.4

    Here's the context-manager::

        @contextmanager
        def redirected_stdio(conn):
            orig_stdin = conn.modules.sys.stdin
            orig_stdout = conn.modules.sys.stdout
            orig_stderr = conn.modules.sys.stderr
            try:
                conn.modules.sys.stdin = sys.stdin
                conn.modules.sys.stdout = sys.stdout
                conn.modules.sys.stderr = sys.stderr
                yield
            finally:
                conn.modules.sys.stdin = orig_stdin
                conn.modules.sys.stdout = orig_stdout
                conn.modules.sys.stderr = orig_stderr

    Example usage::

        with redirected_stdio(conn):
            # remote IO will occur locally

    or ::

        redir = redirected_stdio(conn)
        try:
            # remote IO will occur locally
        finally:
            redir.restore()
    """
    def __init__(self, conn):
        """
        :param conn: the RPyC connection whose stdio will be redirected
        """
        self._restored = True
        self.conn = conn
        self.orig_stdin = self.conn.modules.sys.stdin
        self.orig_stdout = self.conn.modules.sys.stdout
        self.orig_stderr = self.conn.modules.sys.stderr
        self.conn.modules.sys.stdin = sys.stdin
        self.conn.modules.sys.stdout = sys.stdout
        self.conn.modules.sys.stderr = sys.stderr
        self._restored = False
    def __del__(self):
        self.restore()
    def restore(self):
        """Restores the redirection"""
        if self._restored:
            return
        self._restored = True
        self.conn.modules.sys.stdin = self.orig_stdin
        self.conn.modules.sys.stdout = self.orig_stdout
        self.conn.modules.sys.stderr = self.orig_stderr
    def __enter__(self):
        return self
    def __exit__(self, t, v, tb):
        self.restore()


def pm(conn):
    """same as ``pdb.pm()`` but on a remote exception

    :param conn: the RPyC connection
    """
    #pdb.post_mortem(conn.root.getconn()._last_traceback)
    redir = redirected_stdio(conn)
    try:
        conn.modules.pdb.post_mortem(conn.root.getconn()._last_traceback)
    finally:
        redir.restore()

def interact(conn, namespace = None):
    """remote interactive interpreter

    :param conn: the RPyC connection
    :param namespace: the namespace to use (a ``dict``)
    """
    if namespace is None:
        namespace = {}
    namespace["conn"] = conn
    redir = redirected_stdio(conn)
    try:
        conn.execute("""def _rinteract(ns):
            import code
            code.interact(local = dict(ns))""")
        conn.namespace["_rinteract"](namespace)
    finally:
        redir.restore()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2013- Yan Shoshitaishvili aka. zardus
#                     Ruoyu Wang aka. fish
#                     Andrew Dutcher aka. rhelmot
#                     Kevin Borgolte aka. cao
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

__all__ = ["get_memory", "IDAMemory", "CachedIDAMemory",
           "IDAPermissions", "CachedIDAPermissions"]

import collections
import itertools
import logging
import operator
LOG = logging.getLogger("idalink.ida_mem")


# Helper functions.
def _dict_values_sorted_by_key(dictionary):
    # This should be a yield from instead.
    """Internal helper to return the values of a dictionary, sorted by key.
    """
    for _, value in sorted(dictionary.iteritems(), key=operator.itemgetter(0)):
        yield value


def _ondemand(f):
    """Decorator to only request information if not in cache already.
    """
    name = f.__name__

    def func(self, *args, **kwargs):
        if not args and not kwargs:
            if hasattr(self, "_" + name):
                return getattr(self, "_" + name)

            a = f(self, *args, **kwargs)
            setattr(self, "_" + name, a)
            return a
        else:
            return f(self, *args, **kwargs)
    func.__name__ = name
    return func


# Functions others are allowed to call.
def get_memory(idaapi, start, size, default_byte=None):
    # TODO: Documentation
    if idaapi is None:
        idaapi = __import__("idaapi")

    if size == 0:
        return {}

    # We are optimistic and assume it's a continous memory area
    at_address = idaapi.get_many_bytes(start, size)

    d = {}
    if at_address is None:    # It was not, resort to binary research
        if size == 1:
            if default_byte is not None:
                LOG.debug("Using default byte for %d", start)
                d[start] = default_byte
            return d

        mid = start + size / 2
        first_size = mid - start
        second_size = size - first_size

        left = get_memory(idaapi, start, first_size, default_byte=default_byte)
        right = get_memory(idaapi, mid, second_size, default_byte=default_byte)

        if default_byte is None:
            # will be nonsequential
            d.update(left)
            d.update(right)
        else:
            # it will be sequential, so let's combine it
            chained = itertools.chain(_dict_values_sorted_by_key(left),
                                      _dict_values_sorted_by_key(right))
            d[start] = "".join(chained)
    else:
        d[start] = at_address

    return d


class IDAKeys(collections.MutableMapping):  # pylint: disable=W0223
    # TODO: delitem, setitem, getitem are abstract, should be fixed,
    #       disabled warning should be removed
    def __init__(self, ida):
        self.ida = ida

    # Gets the "heads" (instructions and data items) and head sizes from IDA
    @_ondemand
    def heads(self, exclude=()):
        # TODO: Documentation
        LOG.debug("Getting heads from IDA for file %s", self.ida.filename)
        keys = [-1] + list(exclude) + [self.ida.idc.MAXADDR + 1]
        ranges = []
        for i in range(len(keys) - 1):
            a, b = keys[i], keys[i+1]
            if a - b > 1:
                ranges.append((a+1, b-1))

        heads = {}
        for start, end in ranges:
            for head in self.ida.idautils.Heads(start, end, 1):
                heads[head] = self.ida.idc.ItemSize(head)
        return heads

    @_ondemand
    def segments(self):
        # TODO: Documentation
        LOG.debug("Getting segments from IDA for file %s", self.ida.filename)
        segments_size = {}
        for s in self.ida.idautils.Segments():
            segments_size[s] = self.ida.idc.SegEnd(s) - self.ida.idc.SegStart(s)
        return segments_size

    @_ondemand
    def idakeys(self):
        # TODO: Documentation
        keys = set()
        for h, s in self.segments().iteritems():
            for i in range(s):
                keys.add(h + i)
        for h, s in self.heads(exclude=keys).iteritems():
            for i in range(s):
                keys.add(h + i)
        LOG.debug("Done getting keys.")
        return keys

    def __iter__(self):
        # TODO: Refactor to be more pythonic
        for key in self.idakeys():
            yield key

    def __len__(self):
        # This is significantly faster than list(self.__iter__) because
        # we do not need to keep the whole list in memory, just the accumulator.
        return sum(1 for _ in self)

    def __contains__(self, key):
        return key in self.keys()

    def reset(self):
        # TODO: Documentation
        if hasattr(self, "_heads"):
            delattr(self, "_heads")
        if hasattr(self, "_segments"):
            delattr(self, "_segments")
        if hasattr(self, "_idakeys"):
            delattr(self, "_idakeys")


class IDAPermissions(IDAKeys):
    def __init__(self, ida, default_perm=7):
        super(IDAPermissions, self).__init__(ida)
        self.default_perm = default_perm

    def __getitem__(self, address):
        # Only do things that we actually have in IDA
        if address not in self:
            raise KeyError(address)

        seg_start = self.ida.idc.SegStart(address)
        if seg_start == self.ida.idc.BADADDR:
            # We can really only return the default here
            return self.default_perm

        return self.ida.idc.GetSegmentAttr(seg_start, self.ida.idc.SEGATTR_PERM)

    def __setitem__(self, address, value):
        # Nothing we can do here
        pass

    def __delitem__(self, address, value):
        # Nothing we can do here
        pass


class CachedIDAPermissions(IDAPermissions):
    def __init__(self, ida, default_perm=7):
        super(CachedIDAPermissions, self).__init__(ida)
        self.permissions = {}
        self.default_perm = default_perm

    def __getitem__(self, address):
        if address in self.permissions:
            return self.permissions[address]
        p = super(CachedIDAPermissions, self).__getitem__(address)

        # cache the segment
        seg_start = self.ida.idc.SegStart(address)
        seg_end = self.ida.idc.SegEnd(address)
        if seg_start == self.ida.idc.BADADDR:
            self.permissions[address] = p
        else:
            for i in range(seg_start, seg_end):
                self.permissions[i] = p

        return p

    def __setitem__(self, address, value):
        self.permissions[address] = value

    def __delitem__(self, address):
        self.permissions.pop(address, None)

    def reset(self):
        # TODO: Documentation
        self.permissions.clear()
        super(CachedIDAPermissions, self).reset()


class IDAMemory(IDAKeys):
    def __init__(self, ida, default_byte=chr(0xff)):
        super(IDAMemory, self).__init__(ida)
        self.default_byte = default_byte

    def __getitem__(self, address):
        # only do things that we actually have in IDA
        if address not in self:
            raise KeyError(address)

        value = self.ida.idaapi.get_many_bytes(address, 1)
        if value is None:
            value = self.default_byte
        return value

    def __setitem__(self, address, value):
        self.ida.idaapi.patch_byte(address, value)

    def __delitem__(self, address):
        # nothing we can really do here
        pass


class CachedIDAMemory(IDAMemory):
    def __init__(self, ida, default_byte=chr(0xff)):
        super(CachedIDAMemory, self).__init__(ida, default_byte)
        self.local = {}
        self._pulled = False

    @property
    def pulled(self):
        """Check if memory has been pulled from the remote link.
        """
        return self._pulled

    def __getitem__(self, address):
        if address in self.local:
            return self.local[address]

        LOG.debug("Uncached byte: 0x%x", address)
        one = super(CachedIDAMemory, self).__getitem__(address)

        # cache the byte if it's not in a segment
        seg_start = self.ida.idc.SegStart(address)
        if seg_start == self.ida.idc.BADADDR:
            self.local[address] = one
        else:
            # otherwise, cache the segment
            seg_end = self.ida.idc.SegEnd(address)
            seg_size = seg_end - seg_start
            self._load_memory(seg_start, seg_size)

        return one

    def __iter__(self):
        if self.pulled:
            return self.local.__iter__()
        else:
            return super(CachedIDAMemory, self).__iter__()

    def __setitem__(self, address, value):
        self.local[address] = value

    def __delitem__(self, address):
        self.local.pop(address, None)

    def get_memory(self, start, size):
        """Retrieve an area of memory from IDA.
        Returns a sparse dictionary of address -> value.
        """
        LOG.debug("get_memory: %d bytes from %x", size, start)
        return get_memory(self.ida.idaapi, start, size,
                          default_byte=self.default_byte)

    def pull_defined(self):
        if self.pulled:
            return

        start = self.ida.idc.MinEA()
        size = self.ida.idc.MaxEA() - start

        LOG.debug("Loading memory of %s (%d bytes)...", self.ida.filename, size)
        chunks = self.ida.remote_idalink_module.get_memory(None, start, size)

        LOG.debug("Storing loaded memory of %s...", self.ida.filename)
        self._store_loaded_chunks(chunks)

        self._pulled = True

    def reset(self):
        self.local.clear()
        self._pulled = False
        super(CachedIDAMemory, self).reset()

    # Helpers
    def _load_memory(self, start, size):
        chunks = self.get_memory(start, size)
        self.store_loaded_chunks(chunks)

    def _store_loaded_chunks(self, chunks):
        LOG.debug("Updating cache with %d chunks", len(chunks))
        for start, buff in chunks.iteritems():
            for n, i in enumerate(buff):
                if start + n not in self.local:
                    self.local[start + n] = i



#!/usr/bin/env python
"""
classic rpyc server (threaded, forking or std) running a SlaveService
usage:
    rpyc_classic.py                         # default settings
    rpyc_classic.py -m forking -p 12345     # custom settings

    # ssl-authenticated server (keyfile and certfile are required)
    rpyc_classic.py --ssl-keyfile keyfile.pem --ssl-certfile certfile.pem --ssl-cafile cafile.pem
"""
import sys
import os
import rpyc
from optparse import OptionParser
from rpyc.utils.server import ThreadedServer, ForkingServer
from rpyc.utils.classic import DEFAULT_SERVER_PORT, DEFAULT_SERVER_SSL_PORT
from rpyc.utils.registry import REGISTRY_PORT
from rpyc.utils.registry import UDPRegistryClient, TCPRegistryClient
from rpyc.utils.authenticators import SSLAuthenticator
from rpyc.lib import setup_logger
from rpyc.core import SlaveService


parser = OptionParser()
parser.add_option("-m", "--mode", action="store", dest="mode", metavar="MODE",
    default="threaded", type="string", help="mode can be 'threaded', 'forking', "
    "or 'stdio' to operate over the standard IO pipes (for inetd, etc.). "
    "Default is 'threaded'")
#
# TCP options
#
parser.add_option("-p", "--port", action="store", dest="port", type="int",
    metavar="PORT", default=None,
    help="specify a different TCP listener port (default = %s, default for SSL = %s)" %
        (DEFAULT_SERVER_PORT, DEFAULT_SERVER_SSL_PORT))
parser.add_option("--host", action="store", dest="host", type="str",
    metavar="HOST", default="", help="specify a different "
    "host to bind to. Default is INADDR_ANY")
parser.add_option("--ipv6", action="store_true", dest="ipv6",
    metavar="HOST", default=False, help="whether to enable ipv6 or not. " 
    "Default is false")
#
# logging
#
parser.add_option("--logfile", action="store", dest="logfile", type="str",
    metavar="FILE", default=None, help="specify the log file to use; the "
    "default is stderr")
parser.add_option("-q", "--quiet", action="store_true", dest="quiet",
    default=False,
    help="quiet mode (no logging). in stdio mode, writes to /dev/null")
#
# SSL
#
parser.add_option("--ssl-keyfile", action="store", dest="ssl_keyfile", metavar="FILENAME",
    default=None, help="the keyfile to use for SSL. required for SSL"
)
parser.add_option("--ssl-certfile", action="store", dest="ssl_certfile", metavar="FILENAME",
    default=None, help="the certificate file to use for SSL. required for SSL"
)
parser.add_option("--ssl-cafile", action="store", dest="ssl_cafile", metavar="FILENAME",
    default=None, help="the certificate authority chain file to use for SSL. "
    "optional, allows client-side authentication"
)
#
# registry
#
parser.add_option("--register", action="store_true", dest="auto_register",
    default=False, help="asks the server to attempt registering with a registry server"
    "By default, the server will not attempt to register")
parser.add_option("--registry-type", action="store", dest="regtype", type="str",
    default="udp", help="can be 'udp' or 'tcp', default is 'udp'")
parser.add_option("--registry-port", action="store", dest="regport", type="int",
    default=REGISTRY_PORT, help="the UDP/TCP port. default is %s" % (REGISTRY_PORT,))
parser.add_option("--registry-host", action="store", dest="reghost", type="str",
    default=None, help="the registry host machine. for UDP, the default is "
    "255.255.255.255; for TCP, a value is required")


def get_options():
    options, args = parser.parse_args()
    if args:
        parser.error("does not take positional arguments: %r" % (args,))

    options.mode = options.mode.lower()

    if options.regtype.lower() == "udp":
        if options.reghost is None:
            options.reghost = "255.255.255.255"
        options.registrar = UDPRegistryClient(ip = options.reghost, port = options.regport)
    elif options.regtype.lower() == "tcp":
        if options.reghost is None:
            parser.error("must specific --registry-host")
        options.registrar = TCPRegistryClient(ip = options.reghost, port = options.regport)
    else:
        parser.error("invalid registry type %r" % (options.regtype,))

    if options.ssl_keyfile or options.ssl_certfile or options.ssl_cafile:
        if not options.ssl_keyfile:
            parser.error("SSL: keyfile required")
        if not options.ssl_certfile:
            parser.error("SSL: certfile required")
        options.authenticator = SSLAuthenticator(options.ssl_keyfile,
            options.ssl_certfile, options.ssl_cafile)
        if not options.port:
            options.port = DEFAULT_SERVER_SSL_PORT
    else:
        options.authenticator = None
        if not options.port:
            options.port = DEFAULT_SERVER_PORT

    options.handler = "serve_%s" % (options.mode,)
    if options.handler not in globals():
        parser.error("invalid mode %r" % (options.mode,))

    return options

def serve_threaded(options):
    setup_logger(options)
    t = ThreadedServer(SlaveService, hostname = options.host,
        port = options.port, reuse_addr = True, ipv6 = options.ipv6,
        authenticator = options.authenticator, registrar = options.registrar,
        auto_register = options.auto_register)
    t.logger.quiet = options.quiet
    if options.logfile:
        t.logger.console = open(options.logfile, "w")
    t.start()

def serve_forking(options):
    setup_logger(options)
    t = ForkingServer(SlaveService, hostname = options.host,
        port = options.port, reuse_addr = True, ipv6 = options.ipv6,
        authenticator = options.authenticator, registrar = options.registrar,
        auto_register = options.auto_register)
    t.logger.quiet = options.quiet
    if options.logfile:
        t.logger.console = open(options.logfile, "w")
    t.start()

def serve_stdio(options):
    origstdin = sys.stdin
    origstdout = sys.stdout
    if options.quiet:
        dev = os.devnull
    elif sys.platform == "win32":
        dev = "con:"
    else:
        dev = "/dev/tty"
    try:
        sys.stdin = open(dev, "r")
        sys.stdout = open(dev, "w")
    except (IOError, OSError):
        sys.stdin = open(os.devnull, "r")
        sys.stdout = open(os.devnull, "w")
    conn = rpyc.classic.connect_pipes(origstdin, origstdout)
    try:
        try:
            conn.serve_all()
        except KeyboardInterrupt:
            print( "User interrupt!" )
    finally:
        conn.close()


def main():
    options = get_options()
    handler = globals()[options.handler]
    handler(options)


if __name__ == "__main__":
    main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2013- Yan Shoshitaishvili aka. zardus
#                     Ruoyu Wang aka. fish
#                     Andrew Dutcher aka. rhelmot
#                     Kevin Borgolte aka. cao
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# Standard library imports
import logging
import os
import random
import shlex
import socket
import subprocess
import time
import warnings

LOG = logging.getLogger("idalink")

# Local imports
from .rpyc import classic as rpyc_classic
from .memory import CachedIDAMemory, CachedIDAPermissions

# Constants
MODULE_DIR = os.path.dirname(os.path.realpath(__file__))
IDA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "support/")
LOGFILE = "/tmp/idalink-{port}.log"

def which(filename):
    if '/' in filename:
        if os.path.exists(filename) and \
                os.access(filename, os.X_OK):
            return filename
        return None
    path_entries = os.getenv('PATH').split(os.path.pathsep)
    for entry in path_entries:
        filepath = os.path.join(entry, filename)
        if os.path.exists(filepath) and \
                os.access(filepath, os.X_OK):
            return filepath
    return None

# Helper functions
def _ida_spawn(filename, ida_path, port, mode="oneshot",
               processor_type="metapc"):
    """Internal helper function to open IDA on the the file we want to
    analyse.
    """

    ida_realpath = os.path.expanduser(ida_path)
    file_realpath = os.path.realpath(os.path.expanduser(filename))
    logfile = LOGFILE.format(port=port)

    LOG.info("Launching IDA (%s) on %s, listening on port %d, logging to %s",
             ida_realpath, file_realpath, port, logfile)

    # :note: We run IDA through screen because otherwise its UI will hang.
    #        We also setup the environment for IDA.
    #        The other parameters are:
    #        -A     Automatic mode
    #        -S     Run a script (our server script)
    #        -L     Log all output to our logfile
    #        -p     Set the processor type

    command_tpl = "screen -S idalink-{server_port} -d -m " \
                  "'{module_dir}/support/ida_env.sh' '{ida_path}' " \
                  "-M -A "\
                  "-S'{module_dir}/server.py {server_port} {server_mode}' " \
                  "-L'{logfile}' -p{processor} '{file}'"

    command = shlex.split(command_tpl.format(module_dir=MODULE_DIR,
                                             ida_path=ida_realpath,
                                             server_port=port,
                                             server_mode=mode,
                                             logfile=logfile,
                                             processor=processor_type,
                                             file=file_realpath))

    # :note: The above is a bit more Pythonic, if it breaks, use this:
    # screen_name = "idalink-{}".format(port)
    # screen = ["screen", "-S", screen_name, "-d", "-m", "-L", "--"]
    #
    # ida_env = "{}/support/ida_env.sh".format(MODULE_DIR)
    # ida_options = ["-M", "-A",
    #                "-p{}".format(processor_type),
    #                "-S{}/server.py {}".format(MODULE_DIR, port),  # idalink
    #                "-L{}".format(logfile)]                        # logfile
    #
    # ida = [ida_env, ida_realpath]
    # command = screen + ida + ida_options + [file_realpath]

    LOG.debug("IDA command is %s", " ".join(command))
    subprocess.call(command)


def _ida_connect(port):
    link = rpyc_classic.connect("localhost", port)
    LOG.debug("Connected to port %d", port)

    idc = link.root.getmodule("idc")
    idaapi = link.root.getmodule("idaapi")
    idautils = link.root.getmodule("idautils")

    return link, idc, idaapi, idautils


class IDALinkError(Exception):
    pass


class RemoteIDALink(object):
    def __init__(self, filename):
        self.filename = filename
        self.link = None
        self.idc = __import__('idc')
        self.idaapi = __import__('idaapi')
        self.idautils = __import__('idautils')

        self.memory = CachedIDAMemory(self)
        self.permissions = CachedIDAPermissions(self)


class IDALink(object):
    def __init__(self, link, idc, idaapi, idautils, filename=None,
                 pull_memory=True):
        self.filename = filename
        self.link = link
        self.idc = idc
        self.idaapi = idaapi
        self.idautils = idautils

        self.remote_idalink_module = link.root.getmodule("idalink")
        self.remote_link = self.remote_idalink_module.RemoteIDALink(filename)

        self._memory = None
        self.pull_memory = pull_memory
        self._permissions = None

    @property
    def memory(self):
        if self._memory is None:
            self._memory = CachedIDAMemory(self)

            if self.pull_memory:
                self._memory.pull_defined()

        return self._memory

    @memory.deleter
    def memory(self):
        self._memory = None

    @property
    def permissions(self):
        if self._permissions is None:
            self._permissions = CachedIDAPermissions(self)
        return self._permissions

    @permissions.deleter
    def permissions(self):
        self._permissions = None


class idalink(object):
    def __init__(self, filename, ida_prog, retries=60, port=None,
                 spawn=True, pull_memory=True, processor_type="metapc"):
        if port is None:
            port = random.randint(40000, 49999)
        # TODO: check if port is in use

        self._link = None
        self._filename = os.path.realpath(os.path.join(os.getcwd(), filename))
        self._retries = retries
        self._port = port
        self._pull_memory = pull_memory

        progname = which(ida_prog)
        if progname is None:
            raise IDALinkError("Could not find executable %s" % ida_prog)

        if spawn:
            _ida_spawn(self._filename, progname, port, processor_type)

    def __enter__(self):
        for _ in range(self._retries):
            # TODO: detect IDA failure intelligently
            try:
                time.sleep(1)
                LOG.debug("Trying to connect to IDA on port %d", self._port)
                self._link = IDALink(*_ida_connect(self._port),
                                     filename=self._filename,
                                     pull_memory=self._pull_memory)
                break
            except socket.error:
                LOG.debug("... failed. Retrying.")

        if self._link is None:
            raise IDALinkError(("Failed to connect to IDA on port {} for "
                                "file {}").format(self._port, self._filename))

        return self._link

    def __exit__(self, type_, value, traceback):
        try:
            if self._link:
                self._link.idc.Exit(0)
        except EOFError:
            LOG.warning("EOF on link socket, IDA might still be running!")

    @property
    def link(self):
        """Helper property to support the use of idalink without having to
        use a with statement. This property will likely be deprecated and
        might be removed at any point in the future.
        """
        warnings.warn("link property is pending deprecation",
                      PendingDeprecationWarning)
        if self._link is None:
            self.__enter__()
        return self._link



#!/usr/bin/env python
"""
The registry server listens to broadcasts on UDP port 18812, answering to
discovery queries by clients and registering keepalives from all running
servers. In order for clients to use discovery, a registry service must
be running somewhere on their local network.
"""
from optparse import OptionParser
from rpyc.utils.registry import REGISTRY_PORT, DEFAULT_PRUNING_TIMEOUT
from rpyc.utils.registry import UDPRegistryServer, TCPRegistryServer
from rpyc.lib import setup_logger


parser = OptionParser()
parser.add_option("-m", "--mode", action="store", dest="mode", metavar="MODE",
    default="udp", type="string", help="mode can be 'udp' or 'tcp'")
parser.add_option("-p", "--port", action="store", dest="port", type="int",
    metavar="PORT", default=REGISTRY_PORT, help="specify a different UDP/TCP listener port")
parser.add_option("-f", "--file", action="store", dest="logfile", type="str",
    metavar="FILE", default=None, help="specify the log file to use; the default is stderr")
parser.add_option("-q", "--quiet", action="store_true", dest="quiet",
    default=False, help="quiet mode (only errors are logged)")
parser.add_option("-t", "--timeout", action="store", dest="pruning_timeout",
    type="int", default=DEFAULT_PRUNING_TIMEOUT, help="sets a custom pruning timeout")

def main():
    options, args = parser.parse_args()
    if args:
        raise ValueError("does not take positional arguments: %r" % (args,))

    if options.port < 1 or options.port > 65535:
        raise ValueError("invalid TCP/UDP port %r" % (options.port,))

    if options.mode.lower() == "udp":
        server = UDPRegistryServer(port = options.port,
            pruning_timeout = options.pruning_timeout)
    elif options.mode.lower() == "tcp":
        server = TCPRegistryServer(port = options.port,
            pruning_timeout = options.pruning_timeout)
    else:
        raise ValueError("invalid mode %r" % (options.mode,))

    setup_logger(options)
    server.start()


if __name__ == "__main__":
    main()




"""
The RPyC protocol
"""
import sys
import weakref
import itertools
import socket
import time

from threading import Lock
from ..lib.compat import pickle, next, is_py3k, maxint, select_error
from ..lib.colls import WeakValueDict, RefCountingColl
from ..core import consts, brine, vinegar, netref
from .async import AsyncResult

class PingError(Exception):
    """The exception raised should :func:`Connection.ping` fail"""
    pass

DEFAULT_CONFIG = dict(
    # ATTRIBUTES
    allow_safe_attrs = True,
    allow_exposed_attrs = True,
    allow_public_attrs = False,
    allow_all_attrs = False,
    safe_attrs = set(['__abs__', '__add__', '__and__', '__bool__', '__cmp__', '__contains__',
        '__delitem__', '__delslice__', '__div__', '__divmod__', '__doc__',
        '__eq__', '__float__', '__floordiv__', '__ge__', '__getitem__',
        '__getslice__', '__gt__', '__hash__', '__hex__', '__iadd__', '__iand__',
        '__idiv__', '__ifloordiv__', '__ilshift__', '__imod__', '__imul__',
        '__index__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__',
        '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__',
        '__long__', '__lshift__', '__lt__', '__mod__', '__mul__', '__ne__',
        '__neg__', '__new__', '__nonzero__', '__oct__', '__or__', '__pos__',
        '__pow__', '__radd__', '__rand__', '__rdiv__', '__rdivmod__', '__repr__',
        '__rfloordiv__', '__rlshift__', '__rmod__', '__rmul__', '__ror__',
        '__rpow__', '__rrshift__', '__rshift__', '__rsub__', '__rtruediv__',
        '__rxor__', '__setitem__', '__setslice__', '__str__', '__sub__',
        '__truediv__', '__xor__', 'next', '__length_hint__', '__enter__',
        '__exit__', '__next__',]),
    exposed_prefix = "exposed_",
    allow_getattr = True,
    allow_setattr = False,
    allow_delattr = False,
    # EXCEPTIONS
    include_local_traceback = True,
    instantiate_custom_exceptions = False,
    import_custom_exceptions = False,
    instantiate_oldstyle_exceptions = False, # which don't derive from Exception
    propagate_SystemExit_locally = False, # whether to propagate SystemExit locally or to the other party
    log_exceptions = True,
    # MISC
    allow_pickle = False,
    connid = None,
    credentials = None,
    endpoints = None,
    logger = None,
)
"""
The default configuration dictionary of the protocol. You can override these parameters
by passing a different configuration dict to the :class:`Connection` class.

.. note::
   You only need to override the parameters you want to change. There's no need
   to repeat parameters whose values remain unchanged.

=====================================  ================  =====================================================
Parameter                              Default value     Description
=====================================  ================  =====================================================
``allow_safe_attrs``                   ``True``          Whether to allow the use of *safe* attributes
                                                         (only those listed as ``safe_attrs``)
``allow_exposed_attrs``                ``True``          Whether to allow exposed attributes
                                                         (attributes that start with the ``exposed_prefix``)
``allow_public_attrs``                 ``False``         Whether to allow public attributes
                                                         (attributes that don't start with ``_``)
``allow_all_attrs``                    ``False``         Whether to allow all attributes (including private)
``safe_attrs``                         ``set([...])``    The set of attributes considered safe
``exposed_prefix``                     ``"exposed_"``    The prefix of exposed attributes
``allow_getattr``                      ``True``          Whether to allow getting of attributes (``getattr``)
``allow_setattr``                      ``False``         Whether to allow setting of attributes (``setattr``)
``allow_delattr``                      ``False``         Whether to allow deletion of attributes (``delattr``)
``allow_pickle``                       ``False``         Whether to allow the use of ``pickle``

``include_local_traceback``            ``True``          Whether to include the local traceback
                                                         in the remote exception
``instantiate_custom_exceptions``      ``False``         Whether to allow instantiation of
                                                         custom exceptions (not the built in ones)
``import_custom_exceptions``           ``False``         Whether to allow importing of
                                                         exceptions from not-yet-imported modules
``instantiate_oldstyle_exceptions``    ``False``         Whether to allow instantiation of exceptions
                                                         which don't derive from ``Exception``. This
                                                         is not applicable for Python 3 and later.
``propagate_SystemExit_locally``       ``False``         Whether to propagate ``SystemExit``
                                                         locally (kill the server) or to the other
                                                         party (kill the client)
``logger``                             ``None``          The logger instance to use to log exceptions
                                                         (before they are sent to the other party)
                                                         and other events. If ``None``, no logging takes place.

``connid``                             ``None``          **Runtime**: the RPyC connection ID (used
                                                         mainly for debugging purposes)
``credentials``                        ``None``          **Runtime**: the credentails object that was returned
                                                         by the server's :ref:`authenticator <api-authenticators>`
                                                         or ``None``
``endpoints``                          ``None``          **Runtime**: The connection's endpoints. This is a tuple
                                                         made of the local socket endpoint (``getsockname``) and the
                                                         remote one (``getpeername``). This is set by the server
                                                         upon accepting a connection; client side connections
                                                         do no have this configuration option set.
=====================================  ================  =====================================================
"""


_connection_id_generator = itertools.count(1)

class Connection(object):
    """The RPyC *connection* (AKA *protocol*).

    :param service: the :class:`Service <rpyc.core.service.Service>` to expose
    :param channel: the :class:`Channel <rpyc.core.channel.Channel>` over which messages are passed
    :param config: the connection's configuration dict (overriding parameters
                   from the :data:`default configuration <DEFAULT_CONFIG>`)
    :param _lazy: whether or not to initialize the service with the creation of
                  the connection. Default is True. If set to False, you will
                  need to call :func:`_init_service` manually later
    """
    def __init__(self, service, channel, config = {}, _lazy = False):
        self._closed = True
        self._config = DEFAULT_CONFIG.copy()
        self._config.update(config)
        if self._config["connid"] is None:
            self._config["connid"] = "conn%d" % (next(_connection_id_generator),)

        self._channel = channel
        self._seqcounter = itertools.count()
        self._recvlock = Lock()
        self._sendlock = Lock()
        self._sync_replies = {}
        self._async_callbacks = {}
        self._local_objects = RefCountingColl()
        self._last_traceback = None
        self._proxy_cache = WeakValueDict()
        self._netref_classes_cache = {}
        self._remote_root = None
        self._local_root = service(weakref.proxy(self))
        if not _lazy:
            self._init_service()
        self._closed = False
    def _init_service(self):
        self._local_root.on_connect()

    def __del__(self):
        self.close()
    def __enter__(self):
        return self
    def __exit__(self, t, v, tb):
        self.close()
    def __repr__(self):
        a, b = object.__repr__(self).split(" object ")
        return "%s %r object %s" % (a, self._config["connid"], b)

    #
    # IO
    #
    def _cleanup(self, _anyway = True):
        if self._closed and not _anyway:
            return
        self._closed = True
        self._channel.close()
        self._local_root.on_disconnect()
        self._sync_replies.clear()
        self._async_callbacks.clear()
        self._local_objects.clear()
        self._proxy_cache.clear()
        self._netref_classes_cache.clear()
        self._last_traceback = None
        self._remote_root = None
        self._local_root = None
        #self._seqcounter = None
        #self._config.clear()

    def close(self, _catchall = True):
        """closes the connection, releasing all held resources"""
        if self._closed:
            return
        self._closed = True
        try:
            try:
                self._async_request(consts.HANDLE_CLOSE)
            except EOFError:
                pass
            except Exception:
                if not _catchall:
                    raise
        finally:
            self._cleanup(_anyway = True)

    @property
    def closed(self):
        """Indicates whether the connection has been closed or not"""
        return self._closed
    def fileno(self):
        """Returns the connectin's underlying file descriptor"""
        return self._channel.fileno()

    def ping(self, data = None, timeout = 3):
        """
        Asserts that the other party is functioning properly, by making sure
        the *data* is echoed back before the *timeout* expires

        :param data: the data to send (leave ``None`` for the default buffer)
        :param timeout: the maximal time to wait for echo

        :raises: :class:`PingError` if the echoed data does not match
        """
        if data is None:
            data = "abcdefghijklmnopqrstuvwxyz" * 20
        res = self.async_request(consts.HANDLE_PING, data, timeout = timeout)
        if res.value != data:
            raise PingError("echo mismatches sent data")

    def _send(self, msg, seq, args):
        data = brine.dump((msg, seq, args))
        self._sendlock.acquire()
        try:
            self._channel.send(data)
        finally:
            self._sendlock.release()
    def _send_request(self, handler, args):
        seq = next(self._seqcounter)
        self._send(consts.MSG_REQUEST, seq, (handler, self._box(args)))
        return seq
    def _send_reply(self, seq, obj):
        self._send(consts.MSG_REPLY, seq, self._box(obj))
    def _send_exception(self, seq, exctype, excval, exctb):
        exc = vinegar.dump(exctype, excval, exctb,
            include_local_traceback = self._config["include_local_traceback"])
        self._send(consts.MSG_EXCEPTION, seq, exc)

    #
    # boxing
    #
    def _box(self, obj):
        """store a local object in such a way that it could be recreated on
        the remote party either by-value or by-reference"""
        if brine.dumpable(obj):
            return consts.LABEL_VALUE, obj
        if type(obj) is tuple:
            return consts.LABEL_TUPLE, tuple(self._box(item) for item in obj)
        elif isinstance(obj, netref.BaseNetref) and obj.____conn__() is self:
            return consts.LABEL_LOCAL_REF, obj.____oid__
        else:
            self._local_objects.add(obj)
            try:
                cls = obj.__class__
            except Exception:
                # see issue #16
                cls = type(obj)
            return consts.LABEL_REMOTE_REF, (id(obj), cls.__name__, cls.__module__)

    def _unbox(self, package):
        """recreate a local object representation of the remote object: if the
        object is passed by value, just return it; if the object is passed by
        reference, create a netref to it"""
        label, value = package
        if label == consts.LABEL_VALUE:
            return value
        if label == consts.LABEL_TUPLE:
            return tuple(self._unbox(item) for item in value)
        if label == consts.LABEL_LOCAL_REF:
            return self._local_objects[value]
        if label == consts.LABEL_REMOTE_REF:
            oid, clsname, modname = value
            if oid in self._proxy_cache:
                return self._proxy_cache[oid]
            proxy = self._netref_factory(oid, clsname, modname)
            self._proxy_cache[oid] = proxy
            return proxy
        raise ValueError("invalid label %r" % (label,))

    def _netref_factory(self, oid, clsname, modname):
        typeinfo = (clsname, modname)
        if typeinfo in self._netref_classes_cache:
            cls = self._netref_classes_cache[typeinfo]
        elif typeinfo in netref.builtin_classes_cache:
            cls = netref.builtin_classes_cache[typeinfo]
        else:
            info = self.sync_request(consts.HANDLE_INSPECT, oid)
            cls = netref.class_factory(clsname, modname, info)
            self._netref_classes_cache[typeinfo] = cls
        return cls(weakref.ref(self), oid)

    #
    # dispatching
    #
    def _dispatch_request(self, seq, raw_args):
        try:
            handler, args = raw_args
            args = self._unbox(args)
            res = self._HANDLERS[handler](self, *args)
        except KeyboardInterrupt:
            raise
        except:
            # need to catch old style exceptions too
            t, v, tb = sys.exc_info()
            self._last_traceback = tb
            if self._config["logger"]:
                self._config["logger"].debug("Exception caught", exc_info=True)
            if t is SystemExit and self._config["propagate_SystemExit_locally"]:
                raise
            self._send_exception(seq, t, v, tb)
        else:
            self._send_reply(seq, res)

    def _dispatch_reply(self, seq, raw):
        obj = self._unbox(raw)
        if seq in self._async_callbacks:
            self._async_callbacks.pop(seq)(False, obj)
        else:
            self._sync_replies[seq] = (False, obj)

    def _dispatch_exception(self, seq, raw):
        obj = vinegar.load(raw,
            import_custom_exceptions = self._config["import_custom_exceptions"],
            instantiate_custom_exceptions = self._config["instantiate_custom_exceptions"],
            instantiate_oldstyle_exceptions = self._config["instantiate_oldstyle_exceptions"])
        if seq in self._async_callbacks:
            self._async_callbacks.pop(seq)(True, obj)
        else:
            self._sync_replies[seq] = (True, obj)

    #
    # serving
    #
    def _recv(self, timeout, wait_for_lock):
        if not self._recvlock.acquire(wait_for_lock):
            return None
        try:
            try:
                if self._channel.poll(timeout):
                    data = self._channel.recv()
                else:
                    data = None
            except EOFError:
                self.close()
                raise
        finally:
            self._recvlock.release()
        return data

    def _dispatch(self, data):
        msg, seq, args = brine.load(data)
        if msg == consts.MSG_REQUEST:
            self._dispatch_request(seq, args)
        elif msg == consts.MSG_REPLY:
            self._dispatch_reply(seq, args)
        elif msg == consts.MSG_EXCEPTION:
            self._dispatch_exception(seq, args)
        else:
            raise ValueError("invalid message type: %r" % (msg,))

    def poll(self, timeout = 0):
        """Serves a single transaction, should one arrives in the given
        interval. Note that handling a request/reply may trigger nested
        requests, which are all part of a single transaction.

        :returns: ``True`` if a transaction was served, ``False`` otherwise"""
        data = self._recv(timeout, wait_for_lock = False)
        if not data:
            return False
        self._dispatch(data)
        return True

    def serve(self, timeout = 1):
        """Serves a single request or reply that arrives within the given
        time frame (default is 1 sec). Note that the dispatching of a request
        might trigger multiple (nested) requests, thus this function may be
        reentrant.

        :returns: ``True`` if a request or reply were received, ``False``
                  otherwise.
        """
        data = self._recv(timeout, wait_for_lock = True)
        if not data:
            return False
        self._dispatch(data)
        return True

    def serve_all(self):
        """Serves all requests and replies for as long as the connection is
        alive."""
        try:
            try:
                while True:
                    self.serve(0.1)
            except (socket.error, select_error, IOError):
                if not self.closed:
                    raise
            except EOFError:
                pass
        finally:
            self.close()

    def poll_all(self, timeout = 0):
        """Serves all requests and replies that arrive within the given interval.

        :returns: ``True`` if at least a single transaction was served, ``False`` otherwise
        """
        at_least_once = False
        t0 = time.time()
        duration = timeout
        try:
            while True:
                if self.poll(duration):
                    at_least_once = True
                if timeout is not None:
                    duration = t0 + timeout - time.time()
                    if duration < 0:
                        break
        except EOFError:
            pass
        return at_least_once

    #
    # requests
    #
    def sync_request(self, handler, *args):
        """Sends a synchronous request (waits for the reply to arrive)

        :raises: any exception that the requets may be generated
        :returns: the result of the request
        """
        seq = self._send_request(handler, args)
        while seq not in self._sync_replies:
            self.serve(0.1)
        isexc, obj = self._sync_replies.pop(seq)
        if isexc:
            raise obj
        else:
            return obj

    def _async_request(self, handler, args = (), callback = (lambda a, b: None)):
        seq = self._send_request(handler, args)
        self._async_callbacks[seq] = callback
    def async_request(self, handler, *args, **kwargs):
        """Send an asynchronous request (does not wait for it to finish)

        :returns: an :class:`rpyc.core.async.AsyncResult` object, which will
                  eventually hold the result (or exception)
        """
        timeout = kwargs.pop("timeout", None)
        if kwargs:
            raise TypeError("got unexpected keyword argument(s) %s" % (list(kwargs.keys()),))
        res = AsyncResult(weakref.proxy(self))
        self._async_request(handler, args, res)
        if timeout is not None:
            res.set_expiry(timeout)
        return res

    @property
    def root(self):
        """Fetches the root object (service) of the other party"""
        if self._remote_root is None:
            self._remote_root = self.sync_request(consts.HANDLE_GETROOT)
        return self._remote_root

    #
    # attribute access
    #
    def _check_attr(self, obj, name):
        if self._config["allow_exposed_attrs"]:
            if name.startswith(self._config["exposed_prefix"]):
                name2 = name
            else:
                name2 = self._config["exposed_prefix"] + name
            if hasattr(obj, name2):
                return name2
        if self._config["allow_all_attrs"]:
            return name
        if self._config["allow_safe_attrs"] and name in self._config["safe_attrs"]:
            return name
        if self._config["allow_public_attrs"] and not name.startswith("_"):
            return name
        return False

    def _access_attr(self, oid, name, args, overrider, param, default):
        if is_py3k:
            if type(name) is bytes:
                name = str(name, "utf8")
            elif type(name) is not str:
                raise TypeError("name must be a string")
        else:
            if type(name) not in (str, unicode):
                raise TypeError("name must be a string")
            name = str(name) # IronPython issue #10 + py3k issue
        obj = self._local_objects[oid]
        accessor = getattr(type(obj), overrider, None)
        if accessor is None:
            name2 = self._check_attr(obj, name)
            if not self._config[param] or not name2:
                raise AttributeError("cannot access %r" % (name,))
            accessor = default
            name = name2
        return accessor(obj, name, *args)

    #
    # request handlers
    #
    def _handle_ping(self, data):
        return data
    def _handle_close(self):
        self._cleanup()
    def _handle_getroot(self):
        return self._local_root
    def _handle_del(self, oid):
        self._local_objects.decref(oid)
    def _handle_repr(self, oid):
        return repr(self._local_objects[oid])
    def _handle_str(self, oid):
        return str(self._local_objects[oid])
    def _handle_cmp(self, oid, other):
        # cmp() might enter recursive resonance... yet another workaround
        #return cmp(self._local_objects[oid], other)
        obj = self._local_objects[oid]
        try:
            return type(obj).__cmp__(obj, other)
        except (AttributeError, TypeError):
            return NotImplemented
    def _handle_hash(self, oid):
        return hash(self._local_objects[oid])
    def _handle_call(self, oid, args, kwargs=()):
        return self._local_objects[oid](*args, **dict(kwargs))
    def _handle_dir(self, oid):
        return tuple(dir(self._local_objects[oid]))
    def _handle_inspect(self, oid):
        return tuple(netref.inspect_methods(self._local_objects[oid]))
    def _handle_getattr(self, oid, name):
        return self._access_attr(oid, name, (), "_rpyc_getattr", "allow_getattr", getattr)
    def _handle_delattr(self, oid, name):
        return self._access_attr(oid, name, (), "_rpyc_delattr", "allow_delattr", delattr)
    def _handle_setattr(self, oid, name, value):
        return self._access_attr(oid, name, (value,), "_rpyc_setattr", "allow_setattr", setattr)
    def _handle_callattr(self, oid, name, args, kwargs):
        return self._handle_getattr(oid, name)(*args, **dict(kwargs))
    def _handle_pickle(self, oid, proto):
        if not self._config["allow_pickle"]:
            raise ValueError("pickling is disabled")
        return pickle.dumps(self._local_objects[oid], proto)
    def _handle_buffiter(self, oid, count):
        items = []
        obj = self._local_objects[oid]
        i = 0
        try:
            while i < count:
                items.append(next(obj))
                i += 1
        except StopIteration:
            pass
        return tuple(items)
    def _handle_oldslicing(self, oid, attempt, fallback, start, stop, args):
        try:
            # first try __xxxitem__
            getitem = self._handle_getattr(oid, attempt)
            return getitem(slice(start, stop), *args)
        except Exception:
            # fallback to __xxxslice__. see issue #41
            if stop is None:
                stop = maxint
            getslice = self._handle_getattr(oid, fallback)
            return getslice(start, stop, *args)

    # collect handlers
    _HANDLERS = {}
    for name, obj in dict(locals()).items():
        if name.startswith("_handle_"):
            name2 = "HANDLE_" + name[8:].upper()
            if hasattr(consts, name2):
                _HANDLERS[getattr(consts, name2)] = obj
            else:
                raise NameError("no constant defined for %r", name)
    del name, name2, obj




from .stream import SocketStream, TunneledSocketStream, PipeStream
from .channel import Channel
from .protocol import Connection
from .netref import BaseNetref
from .async import AsyncResult, AsyncResultTimeout
from .service import Service, VoidService, SlaveService
from .vinegar import GenericException



import os
from subprocess import Popen, PIPE
from rpyc.lib import safe_import
from rpyc.lib.compat import BYTES_LITERAL
signal = safe_import("signal")

# modified from the stdlib pipes module for windows
_safechars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@%_-+=:,./'
_funnychars = '"`$\\'
def shquote(text):
    if not text:
        return "''"
    for c in text:
        if c not in _safechars:
            break
    else:
        return text
    if "'" not in text:
        return "'" + text + "'"
    def escaped(c):
        if c in _funnychars:
            return '\\' + c 
        else:
            return c
    res = "".join(escaped(c) for c in text)
    return '"' + res + '"'

class ProcessExecutionError(Exception):
    """raised by :func:`SshContext.execute` should the executed process 
    terminate with an error"""
    pass

import subprocess
def _get_startupinfo():
    if subprocess.mswindows:
        import _subprocess
        sui = subprocess.STARTUPINFO()
        sui.dwFlags |= _subprocess.STARTF_USESHOWWINDOW #@UndefinedVariable
        sui.wShowWindow = _subprocess.SW_HIDE #@UndefinedVariable
        return sui
    else:
        return None

class SshTunnel(object):
    """
    Represents an active SSH tunnel (as created by ``ssh -L``).
    
    .. note:: 
       Do not instantiate this class yourself -- use the :func:`SshContext.tunnel`
       function for that.
    """
    
    PROGRAM = r"""import sys;sys.stdout.write("ready\n\n\n");sys.stdout.flush();sys.stdin.readline()"""

    def __init__(self, sshctx, loc_host, loc_port, rem_host, rem_port):
        self.loc_host = loc_host
        self.loc_port = loc_port
        self.rem_host = rem_host
        self.rem_port = rem_port
        self.sshctx = sshctx
        self.proc = sshctx.popen("python", "-u", "-c", self.PROGRAM,
            L = "[%s]:%s:[%s]:%s" % (loc_host, loc_port, rem_host, rem_port))
        banner = self.proc.stdout.readline().strip()
        if banner != BYTES_LITERAL("ready"):
            raise ValueError("tunnel failed", banner)
    def __del__(self):
        try:
            self.close()
        except Exception:
            pass
    def __str__(self):
        return "%s:%s --> (%s)%s:%s" % (self.loc_host, self.loc_port, self.sshctx.host,
            self.rem_host, self.rem_port)
    def is_open(self):
        """returns True if the ``ssh`` process is alive, False otherwise"""
        return self.proc and self.proc.poll() is None
    def close(self):
        """closes (terminates) the SSH tunnel"""
        if not self.is_open():
            return
        self.proc.stdin.write(BYTES_LITERAL("foo\n\n\n"))
        self.proc.stdin.close()
        self.proc.stdout.close()
        self.proc.stderr.close()
        try:
            self.proc.kill()
        except AttributeError:
            if signal:
                os.kill(self.proc.pid, signal.SIGTERM)
        self.proc.wait()
        self.proc = None

class SshContext(object):
    """
    An *SSH context* encapsulates all the details required to establish an SSH 
    connection to other host. It includes the host name, user name, TCP port, 
    identity file, etc.
    
    Once constructed, it can serve as a factory for SSH operations, such as 
    executing a remote program and getting its stdout, or uploading/downloading
    files using ``scp``. It also serves for creating SSH tunnels. 
    
    Example::
    
        >>> sshctx = SshContext("mymachine", username="borg", keyfile="/home/foo/.ssh/mymachine-id")
        >>> sshctx.execute("ls")
        (0, "...", "")
    """
    def __init__(self, host, user = None, port = None, keyfile = None,
            ssh_program = "ssh", ssh_env = None, ssh_cwd = None,
            scp_program = "scp", scp_env = None, scp_cwd = None):
        self.host = host
        self.user = user
        self.port = port
        self.keyfile = keyfile
        self.ssh_program = ssh_program
        self.ssh_env = ssh_env
        self.ssh_cwd = ssh_cwd
        self.scp_program = scp_program
        self.scp_env = scp_env
        self.scp_cwd = scp_cwd

    def __str__(self):
        uri = "ssh://"
        if self.user:
            uri += "%s@%s" % (self.user, self.host)
        else:
            uri += self.host
        if self.port:
            uri += ":%d" % (self.port)
        return uri

    def _convert_kwargs_to_args(self, kwargs):
        args = []
        for k, v in kwargs.items():
            if v is True:
                args.append("-%s" % (k,))
            elif v is False:
                pass
            else:
                args.append("-%s" % (k,))
                args.append(str(v))
        return args

    def _process_scp_cmdline(self, kwargs):
        args = [self.scp_program]
        if "r" not in kwargs:
            kwargs["r"] = True
        if self.keyfile and "i" not in kwargs:
            kwargs["i"] = self.keyfile
        if self.port and "P" not in kwargs:
            kwargs["P"] = self.port
        args.extend(self._convert_kwargs_to_args(kwargs))
        if self.user:
            host = "%s@%s" % (self.user, self.host)
        else:
            host = self.host
        return args, host

    def _process_ssh_cmdline(self, kwargs):
        args = [self.ssh_program]
        if self.keyfile and "i" not in kwargs:
            kwargs["i"] = self.keyfile
        if self.port and "p" not in kwargs:
            kwargs["p"] = self.port
        args.extend(self._convert_kwargs_to_args(kwargs))
        if self.user:
            args.append("%s@%s" % (self.user, self.host))
        else:
            args.append(self.host)
        return args

    def popen(self, *args, **kwargs):
        """Runs the given command line remotely (over SSH), returning the 
        ``subprocess.Popen`` instance of the command
        
        :param args: the command line arguments
        :param kwargs: additional keyword arguments passed to ``ssh``
        
        :returns: a ``Popen`` instance
        
        Example::
            
            proc = ctx.popen("ls", "-la")
            proc.wait()
        """
        cmdline = self._process_ssh_cmdline(kwargs)
        cmdline.extend(shquote(a) for a in args)
        return Popen(cmdline, stdin = PIPE, stdout = PIPE, stderr = PIPE,
            cwd = self.ssh_cwd, env = self.ssh_env, shell = False, 
            startupinfo = _get_startupinfo())

    def execute(self, *args, **kwargs):
        """Runs the given command line remotely (over SSH), waits for it to finish,
        returning the return code, stdout, and stderr of the executed process.
        
        :param args: the command line arguments
        :param kwargs: additional keyword arguments passed to ``ssh``, except for
                       ``retcode`` and ``input``.
        :param retcode: *keyword only*, the expected return code (Defaults to 0 
                        -- success). An exception is raised if the return code does
                        not match the expected one, unless it is ``None``, in 
                        which case it will not be tested.
        :param input: *keyword only*, an input string that will be passed to 
                      ``Popen.communicate``. Defaults to ``None``
        
        :raises: :class:`ProcessExecutionError` if the expected return code 
                 is not matched
        
        :returns: a tuple of (return code, stdout, stderr)
        
        Example::
            
            rc, out, err = ctx.execute("ls", "-la")
        """
        retcode = kwargs.pop("retcode", 0)
        input = kwargs.pop("input", None)
        proc = self.popen(*args, **kwargs)
        stdout, stderr = proc.communicate(input)
        if retcode is not None and proc.returncode != retcode:
            raise ProcessExecutionError(proc.returncode, stdout, stderr)
        return proc.returncode, stdout, stderr

    def upload(self, src, dst, **kwargs):
        """
        Uploads *src* from the local machine to *dst* on the other side. By default, 
        ``-r`` (recursive copy) is given to ``scp``, so *src* can be either a file or 
        a directory. To override this behavior, pass ``r = False`` as a keyword argument. 
        
        :param src: the source path (on the local side)
        :param dst: the destination path (on the remote side)
        :param kwargs: any additional keyword arguments, passed to ``scp``.
        """
        cmdline, host = self._process_scp_cmdline(kwargs)
        cmdline.append(src)
        cmdline.append("%s:%s" % (host, dst))
        proc = Popen(cmdline, stdin = PIPE, stdout = PIPE, stderr = PIPE, shell = False,
            cwd = self.scp_cwd, env = self.scp_env, startupinfo = _get_startupinfo())
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            raise ValueError("upload failed", stdout, stderr)

    def download(self, src, dst, **kwargs):
        """
        Downloads *src* from the other side to *dst* on the local side. By default, 
        ``-r`` (recursive copy) is given to ``scp``, so *src* can be either a file or 
        a directory. To override this behavior, pass ``r = False`` as a keyword argument.
        
        :param src: the source path (on the other side)
        :param dst: the destination path (on the local side)
        :param kwargs: any additional keyword arguments, passed to ``scp``.
        """
        cmdline, host = self._process_scp_cmdline(kwargs)
        cmdline.append("%s:%s" % (host, src))
        cmdline.append(dst)
        proc = Popen(cmdline, stdin = PIPE, stdout = PIPE, stderr = PIPE, shell = False,
            cwd = self.scp_cwd, env = self.scp_env, startupinfo = _get_startupinfo())
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            raise ValueError("upload failed", stdout, stderr)

    def tunnel(self, loc_port, rem_port, loc_host = "localhost", rem_host = "localhost"):
        """
        Creates an SSH tunnel from the local port to the remote one. This is
        translated to ``ssh -L loc_host:loc_port:rem_host:rem_port``.
        
        :param loc_port: the local TCP port to forward
        :param rem_port: the remote (server) TCP port, to which the local port 
                         will be forwarded
        
        :returns: an :class:`SshTunnel` instance
        """
        return SshTunnel(self, loc_host, loc_port, rem_host, rem_port)




from distutils.core import setup
setup(
    name = 'idalink',
    description = 'An interface to the insides of IDA!',
    version = '0.10',
    packages = ['idalink', 'idalink.rpyc', 'idalink.rpyc.core', 'idalink.rpyc.scripts', 'idalink.rpyc.utils', 'idalink.rpyc.lib'],
    package_data = { 'idalink': [ 'support/*.sh' ] },
    install_requires=['rpyc']
)



"""
Helpers and wrappers for common RPyC tasks
"""
import time
import threading
from ..lib.colls import WeakValueDict
from ..lib.compat import callable
from ..core.consts import HANDLE_BUFFITER, HANDLE_CALL
from ..core.netref import syncreq, asyncreq


def buffiter(obj, chunk = 10, max_chunk = 1000, factor = 2):
    """Buffered iterator - reads the remote iterator in chunks starting with
    *chunk*, multiplying the chunk size by *factor* every time, as an
    exponential-backoff, up to a chunk of *max_chunk* size.

    ``buffiter`` is very useful for tight loops, where you fetch an element
    from the other side with every iterator. Instead of being limited by the
    network's latency after every iteration, ``buffiter`` fetches a "chunk"
    of elements every time, reducing the amount of network I/Os.

    :param obj: An iterable object (supports ``iter()``)
    :param chunk: the initial chunk size
    :param max_chunk: the maximal chunk size
    :param factor: the factor by which to multiply the chunk size after every
                   iterator (up to *max_chunk*). Must be >= 1.

    :returns: an iterator

    Example::

        cursor = db.get_cursor()
        for id, name, dob in buffiter(cursor.select("Id", "Name", "DoB")):
            print id, name, dob
    """
    if factor < 1:
        raise ValueError("factor must be >= 1, got %r" % (factor,))
    it = iter(obj)
    count = chunk
    while True:
        items = syncreq(it, HANDLE_BUFFITER, count)
        count = min(count * factor, max_chunk)
        if not items:
            break
        for elem in items:
            yield elem

def restricted(obj, attrs, wattrs = None):
    """Returns a 'restricted' version of an object, i.e., allowing access only to a subset of its
    attributes. This is useful when returning a "broad" or "dangerous" object, where you don't
    want the other party to have access to all of its attributes.

    .. versionadded:: 3.2

    :param obj: any object
    :param attrs: the set of attributes exposed for reading (``getattr``) or writing (``setattr``).
                  The same set will serve both for reading and writing, unless wattrs is explicitly
                  given.
    :param wattrs: the set of attributes exposed for writing (``setattr``). If ``None``,
                   ``wattrs`` will default to ``attrs``. To disable setting attributes completely,
                   set to an empty tuple ``()``.
    :returns: a restricted view of the object

    Example::

        class MyService(rpyc.Service):
            def exposed_open(self, filename):
                f = open(filename, "r")
                return rpyc.restricted(f, {"read", "close"})   # disallow access to `seek` or `write`

    """
    if wattrs is None:
        wattrs = attrs
    class Restricted(object):
        def _rpyc_getattr(self, name):
            if name not in attrs:
                raise AttributeError(name)
            return getattr(obj, name)
        __getattr__ = _rpyc_getattr
        def _rpyc_setattr(self, name, value):
            if name not in wattrs:
                raise AttributeError(name)
            setattr(obj, name, value)
        __setattr__ = _rpyc_setattr
    return Restricted()

class _Async(object):
    """Creates an async proxy wrapper over an existing proxy. Async proxies
    are cached. Invoking an async proxy will return an AsyncResult instead of
    blocking"""

    __slots__ = ("proxy", "__weakref__")
    def __init__(self, proxy):
        self.proxy = proxy
    def __call__(self, *args, **kwargs):
        return asyncreq(self.proxy, HANDLE_CALL, args, tuple(kwargs.items()))
    def __repr__(self):
        return "async(%r)" % (self.proxy,)

_async_proxies_cache = WeakValueDict()
def async(proxy):
    """
    Returns an asynchronous "version" of the given proxy. Invoking the returned
    proxy will not block; instead it will return an
    :class:`rpyc.core.async.AsyncResult` object that you can test for completion

    :param proxy: any **callable** RPyC proxy

    :returns: the proxy, wrapped by an asynchronous wrapper

    Example::

        async_sleep = rpyc.async(conn.modules.time.sleep)
        res = async_sleep(5)

    .. _async_note:

    .. note::
       In order to avoid overloading the GC, the returned asynchronous wrapper is
       cached as a weak reference. Therefore, do not use::

           rpyc.async(foo)(5)

       Always store the returned asynchronous wrapper in a variable, e.g. ::

           a_foo = rpyc.async(foo)
           a_foo(5)
    """
    pid = id(proxy)
    if pid in _async_proxies_cache:
        return _async_proxies_cache[pid]
    if not hasattr(proxy, "____conn__") or not hasattr(proxy, "____oid__"):
        raise TypeError("'proxy' must be a Netref: %r", (proxy,))
    if not callable(proxy):
        raise TypeError("'proxy' must be callable: %r" % (proxy,))
    caller = _Async(proxy)
    _async_proxies_cache[id(caller)] = _async_proxies_cache[pid] = caller
    return caller

async.__doc__ = _Async.__doc__

class timed(object):
    """Creates a timed asynchronous proxy. Invoking the timed proxy will
    run in the background and will raise an :class:`rpyc.core.async.AsyncResultTimeout`
    exception if the computation does not terminate within the given time frame

    :param proxy: any **callable** RPyC proxy
    :param timeout: the maximal number of seconds to allow the operation to run

    :returns: a ``timed`` wrapped proxy

    Example::

        t_sleep = rpyc.timed(conn.modules.time.sleep, 6) # allow up to 6 seconds
        t_sleep(4) # okay
        t_sleep(8) # will time out and raise AsyncResultTimeout
    """

    __slots__ = ("__weakref__", "proxy", "timeout")
    def __init__(self, proxy, timeout):
        self.proxy = async(proxy)
        self.timeout = timeout
    def __call__(self, *args, **kwargs):
        res = self.proxy(*args, **kwargs)
        res.set_expiry(self.timeout)
        return res
    def __repr__(self):
        return "timed(%r, %r)" % (self.proxy.proxy, self.timeout)

class BgServingThread(object):
    """Runs an RPyC server in the background to serve all requests and replies
    that arrive on the given RPyC connection. The thread is started upon the
    the instantiation of the ``BgServingThread`` object; you can use the
    :meth:`stop` method to stop the server thread

    Example::

        conn = rpyc.connect(...)
        bg_server = BgServingThread(conn)
        ...
        bg_server.stop()

    .. note::
       For a more detailed explanation of asynchronous operation and the role of the
       ``BgServingThread``, see :ref:`tut5`

    """
    # these numbers are magical...
    SERVE_INTERVAL = 0.0
    SLEEP_INTERVAL = 0.1

    def __init__(self, conn):
        self._conn = conn
        self._thread = threading.Thread(target = self._bg_server)
        self._thread.setDaemon(True)
        self._active = True
        self._thread.start()
    def __del__(self):
        if self._active:
            self.stop()
    def _bg_server(self):
        try:
            while self._active:
                self._conn.serve(self.SERVE_INTERVAL)
                time.sleep(self.SLEEP_INTERVAL) # to reduce contention
        except Exception:
            if self._active:
                self._active = False
                raise
    def stop(self):
        """stop the server thread. once stopped, it cannot be resumed. you will
        have to create a new BgServingThread object later."""
        assert self._active
        self._active = False
        self._thread.join()
        self._conn = None




import weakref
from threading import Lock

class WeakValueDict(object):
    """a light-weight version of weakref.WeakValueDictionary"""
    __slots__ = ("_dict",)
    def __init__(self):
        self._dict = {}
    def __repr__(self):
        return repr(self._dict)
    def __iter__(self):
        return self.iterkeys()
    def __len__(self):
        return len(self._dict)
    def __contains__(self, key):
        try:
            self[key]
        except KeyError:
            return False
        else:
            return True
    def get(self, key, default = None):
        try:
            return self[key]
        except KeyError:
            return default
    def __getitem__(self, key):
        obj = self._dict[key]()
        if obj is None:
            raise KeyError(key)
        return obj
    def __setitem__(self, key, value):
        def remover(wr, _dict = self._dict, key = key):
            _dict.pop(key, None)
        self._dict[key] = weakref.ref(value, remover)
    def __delitem__(self, key):
        del self._dict[key]
    def iterkeys(self):
        return self._dict.keys()
    def keys(self):
        return self._dict.keys()
    def itervalues(self):
        for k in self:
            yield self[k]
    def values(self):
        return list(self.itervalues())
    def iteritems(self):
        for k in self:
            yield k, self[k]
    def items(self):
        return list(self.iteritems())
    def clear(self):
        self._dict.clear()

class RefCountingColl(object):
    """a set-like object that implements refcounting on its contained objects"""
    __slots__ = ("_lock", "_dict")
    def __init__(self):
        self._lock = Lock()
        self._dict = {}
    def __repr__(self):
        return repr(self._dict)
    def add(self, obj):
        self._lock.acquire()
        try:
            key = id(obj)
            slot = self._dict.get(key, None)
            if slot is None:
                slot = [obj, 0]
            else:
                slot[1] += 1
            self._dict[key] = slot
        finally:
            self._lock.release()
    def clear(self):
        self._lock.acquire()
        try:
            self._dict.clear()
        finally:
            self._lock.release()
    def decref(self, key):
        self._lock.acquire()
        try:
            slot = self._dict[key]
            if slot[1] < 1:
                del self._dict[key]
            else:
                slot[1] -= 1
                self._dict[key] = slot
        finally:
            self._lock.release()
    def __getitem__(self, key):
        self._lock.acquire()
        try:
            return self._dict[key][0]
        finally:
            self._lock.release()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2013- Yan Shoshitaishvili aka. zardus
#                     Ruoyu Wang aka. fish
#                     Andrew Dutcher aka. rhelmot
#                     Kevin Borgolte aka. cao
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import threading

# idc is just within IDA, so make pylint stop complaining
import idc      # pylint: disable=F0401

# pylint: disable=W0403
# :note: Those should be relative imports, but IDA doesn't like them.
from rpyc.core import SlaveService
from rpyc.utils.server import OneShotServer, ThreadedServer

if __name__ == "__main__":
    print "Received arguments: {}".format(idc.ARGV)

    port = int(idc.ARGV[1]) if idc.ARGV[1:] else 18861
    mode = idc.ARGV[2] if idc.ARGV[2:] else "oneshot"

    # :note: For speed, we don't want to idc.Wait() here,
    #        but you might want to call it in your code
    #        to make sure that autoanalysis has finished.

    if mode == "threaded":
        ThreadedServer(SlaveService, port=port).start()
    else:
        OneShotServer(SlaveService, port=port).start()
        idc.Exit(0)
